#!/usr/bin/python
#
# Copyright 2012 Canonical Ltd
# Authors:
#   Kyle Nitzsche <kyle.nitzsche@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import optparse
import sys
import subprocess

def dot2png(_f, shape="box"):
    """Run dot on passed file handle to create png with box shapes as
    default"""

    _fpng = _f.rstrip(".dot") + ".png"
    cmd = ['dot', '-Tpng', '-Nshape=%s' % shape, '-o', _fpng, _f]
    subprocess.call(cmd)

def replacement(p):
    """Fix pkg names such that they are valid dotty node names"""

    #print "before: %s" % p,
    p = p.replace('-','_')
    p = p.replace('-','_')
    p = p.replace(':','_')
    p = p.replace('.','_')
    p = p.replace('+','Plus')
    #print " after: %s" % p
    return p

class Dot():
    """Provides an API for extracting information from and manipulating
    a .dot file. Dot file must not be pretty printed, that is, all node
    data must be on a single line and all join ('edge') info must be on
    a single line.
    """

    def __init__(self, path, shape="box"):
        """Reads dot file and sets up object variables with appropriate parts
        of the dot file, including the header, content, nodes, joins, footer
       lnodes, and rnodes.
        """

        self.path = path
        try:
            f = open(self.path, 'r')
        except IOError, e:
            print e
            print "Stopping."
            sys.exit(1)
        self.dot = f.readlines()
        f.close()
        self.root = self.get_root()
        self.header = self.get_header()
        self.content = self.get_content()
        self.nodes = self.get_nodes()
        self.joins = self.get_joins()
        self.lnodes = self.get_lnodes()
        self.rnodes = self.get_rnodes()
        self.footer = "}\n}"
        self.shape = shape

    def get_root(self):
        line = self.dot[0]
        if line.startswith("//root:"):
            root = line[line.find(':')+1:].rstrip()
            return root
        else:
            return ""

    def get_header(self):
        """Return the header lines of the dot file as a list of lines"""
        header = []
        for line in self.dot:
            if (line.find("{") > -1 
                or line.startswith("label")
                or line.startswith("//")):
                header.append(line)
            else: break
        return header

    def get_content(self):
        """Returns the content of the dot file as a list of lines.

        Content is the dot file without the header or footer.
        """
        content = []
        for line in self.dot:
            if line.find("}") > -1: continue
            if line.find("{") > -1: continue
            if line.startswith("label"): continue
            content.append(line)
        #print content
        return content

    def get_joins(self):
        """Returns the 'join' lines from the dot file as a list of lists [[]].

        Join lines are edges between nodes, specifically: lnode -> rnode.
        There may be a style after the rnode: lnode -> rnode style. Each
        top-level list represents a join line. This contains a list containing
        the items in the line as follows: [lnode, rnode, style(if any)]
        """

        joins = []
        # joins structure: [lnode:[rnode,style]]
        for line in self.content:
            line = line.strip()
            if len(line) == 0: continue
            if line.find("->") > -1: # so, is not node line but a join
                tokens = line.split(" ")
                if tokens in joins: continue
                if len(tokens) < 1: continue
                if len(tokens) == 3:
                    joins.append([tokens[0],tokens[2].rstrip(";\n"), ""])
                if len(tokens) == 4:
                    joins.append(
                        [tokens[0],
                        tokens[2],
                        tokens[3].rstrip(";\n")]
                                )
        return joins

    def get_nodes(self):
        """Returns a dictionary of nodes in the dot file where the key is the
        node name and the value is the style, if any, as a string, without the
        enclosing square brackets one finds in a dot file.
        """

        nodes = dict()
        node_lines = []
        for line in self.content:
            if line.find('->') == -1: #so is not a join
                tokens = line.split()
                node = tokens[0]
                if len(tokens) > 1: # tokens[1] is a style
                    style = tokens[1][1:tokens[1].find(']')]
                else:
                    style=""
                if node not in nodes.keys():
                    nodes[node] = style
        return nodes

    def get_lnodes(self):
        """Returns a dictionary of left hand nodes ('lnodes') from self.joins
        where the key is lnode name and the value is the number of times the
        lnode exists as an lnode in the dot file.
        """

        lnodes = dict()
        for join in self.joins:
            if len(join) == 0: continue
            node = join[0]
            if node in lnodes.keys():
                lnodes[node] += 1
            else:
                lnodes[node] = 1
        return lnodes

    def get_rnodes(self):
        """Returns a dictionary of right hand nodes ('rnodes') from self.joins
        where the key is rnode name and the value is the number of times the
        rnode exists as an rnode in the dot file.
        """

        rnodes = dict()
        for join in self.joins:
            if len(join) == 0: continue
            node = join[1]
            if node in rnodes.keys():
                rnodes[node] += 1
            else:
                rnodes[node] = 1
        return rnodes

    def get_popular_lnodes(self, popularity):
        """Returns a dictionary of lnodes where all lnodes exiist as lnodes in
        the dot at least as many times as specified by the 'popularity' arg.
        """
        popular = dict()
        for node, num in self.lnodes.iteritems():
            if int(num) >= int(popularity):
                popular[node] = num
        return popular

    def get_rnodes_for_lnode(self, lnode):
        """Returns a list of all rnodes of the passed lnode"""

        rnodes = []
        for join in self.joins:
            _lnode = join[0]
            if _lnode != lnode: continue
            rnode = join[1]
            if rnode not in rnodes:
                rnodes.append(rnode)
        return rnodes

    def get_lnodes_for_rnode(self,rnode):
        """Returns a list of all lnodes of the passed rnode"""

        lnodes = []
        for join in self.joins:
            _rnode = join[1]
            if _rnode != rnode: continue
            lnode = join[0]
            if lnode not in lnodes:
                lnodes.append(lnode)
        return lnodes

    def get_fnodes(self, node, fnodes, other_nodes):
        """Used recursively to find node sets that are summarizable as a
        family in order to make large graphs smaller and easier to comprehend.

        A family of nodes is a set of nodes that have no lnodes outside of the
        family (except for the root node of the family which does have lnodes
        outside of the family).
        """

        rnodes = self.get_rnodes_for_lnode(node)
        for rnode in rnodes:
            rnode_valid = True
            lnodes = self.get_lnodes_for_rnode(rnode)
            for lnode in lnodes:
                if lnode in other_nodes:
                    rnode_valid = False
                    break
            if rnode not in fnodes:
                if not rnode_valid: continue
                fnodes.append(rnode)
                if rnode in other_nodes:
                    other_nodes.remove(rnode)
                fnodes = self.get_fnodes(rnode, fnodes, other_nodes)
        return fnodes

    def get_families(self):
        """Returns a dictionary of families where the key is the family's root
        node name and the value is the list of nodes in the family.
        """

        fam = dict()
        fams = dict()

        for node in self.nodes.keys():
            onodes = list(self.nodes.keys())
            onodes.remove(node)
            n_nodes = dict()
            fam[node] = self.get_fnodes(node, [node], onodes)
            if len(fam[node]) < 2:
                #if opt.verbose:
                    #print "%s has no family." % node
                continue
            fams[node] = fam[node]
            #print fams[node]
        return fams

    def remove_root_family(self, fams, root):
        """If the root node is found to be summarizable as a family, remove it
        from the dictionary of families in order to keep the gerated data
        clear. That is, all direct depends of the root node should be separate
        nodes rather than being included in the top level famliy.
        """

        if root == "":
            print "Error. You must set the root package (node) with -r PKG.\
                Stopping"
            sys.exit(1)
        try:
            fams.pop(root)
        except KeyError:
            pass # it's OK if the root node is not a family
        return fams

    def process_families(self,fams):
        """Returns a dictionary made from the passed fams that has a key for
        each node the value of which is a tuple of the family it is in and the
        size of the family.
        """

        members = dict()

        for fam in fams:
            size = len(fams[fam])
            for _pkg in fams[fam]:
                members[_pkg] = (fam, size)
        #print members
        return members

    def process_joins(self, fams, fam):
        """Returns processed joins such that lnodes and rnodes are replaced
        with the family they are in, if any, weith a proper family name
        (pkg_Family_size).

        This is the key step that reduces the complexity of the graph by
        summarizing nodes that are families as single family nodes in the
        generated dot file.

        Note that the dot style for the join is also added, if it exists.
        """

        n_joins = []
        fams = self.process_families(fams)
        for join in self.joins:
            _join = []
            lnode = join[0]
            rnode = join[1]
            if lnode in fams:
                lnode = "%s_Family_%d" %(fams[lnode][0],fams[lnode][1])
            elif rnode in fams:
                rnode = "%s_Family_%d" %(fams[rnode][0],fams[rnode][1])
            _join.append(lnode)
            _join.append(rnode)
            if len(join) > 2:
                _join.append(join[2])
            if lnode == rnode: continue # remove circular joins
            if _join not in n_joins:
                n_joins.append(_join)
        return n_joins

    def process_nodes(self, fams, fam, f_internal_nodes):
        """Returns a pruned dictionary of nodes suitable for output to dot
        where the key is node and the value is an approriately amended style.

        Nodes internal to families that are not the family top node are
        dropped as part of family summarization.
        """

        n_nodes = dict()
        for node in self.nodes.keys():
            if node in f_internal_nodes:
                continue
            style = self.nodes[node]
            if node == fam:
                styles = style.split(',')
                #print "node %s inode %s styles %s" % (node, inode, styles)
                n_styles = ''
                for _s in styles:
                    if not _s.startswith("label="):
                        n_styles+= _s + ','
                if style.find('fillcolor=blue') >-1:#this is the root node
                    style = n_styles
                    style += "fillcolor=blue,style=filled,shape="
                    style += self.shape
                else:
                    style = n_styles
                    style += "fillcolor=yellow,style=filled,shape="
                    style += self.shape
                node = "%s_Family_%d" % (node, len(fams[fam]))
            if node != fam and node in fams[fam]:
                continue
            n_nodes[node] = style
        return n_nodes

    def remove_subset_families(self, fams):
        """Returns the passed fams dictionary with 'subset' families removed.

        That is, one famiily may fully contian a smaller family. In such
        cases, the smaller family should be deleted since the goal is to
        provide only the largest families. Thus, subset families are pruned.
        """

        _fams = dict(fams)
        for fam in fams:
            idx = -1
            for pkg in fams[fam]:
                idx +=1
                if idx == 0: continue
                try:
                    _fams.pop(pkg)
                except:
                    pass
        return _fams

    def summarize(self, root=None, out_file="out", dont_run_dot=False):
        """The driver method for family summarizatio.

        Finds families and summarizes them as nodes.

        Prunes and processes nodes and joins.

        Creates dots and converts them to png unless directed not to.

        Creates FAMILIES json and writes to file.
        """

        fams_with_root = self.get_families()
        fams = self.remove_root_family(fams_with_root, root)

        # gets family internal nodes, not including top family fnode
        # so that later such nodes can be dropped from nodes in dot
        f_internal_nodes = []

        for fam in fams:
            for n in fams[fam]:
                if n not in f_internal_nodes and n != fam:
                    f_internal_nodes.append(n)
                    #print "internal %s " % n

        fams = self.remove_subset_families(fams)

        for fam in fams:
            #print "FAM %s fams[fam] %s" % (fam, fams[fam])
            self.joins = self.process_joins(fams, fam)
            self.nodes = self.process_nodes(fams, fam, f_internal_nodes)

        #print "2 len(fams): %d" % len(fams)
        nodes = list(set(self.nodes)) # remove redundants
        _joins = []
        for i in self.joins: #remove redundants
            if i not in _joins:
                _joins.append(i)
        joins = _joins

        # make and  print dot file out
        res = ""
        for line in self.header:
            res += line
        for node in self.nodes:
            res += node + " [" + self.nodes[node] + "];\n"
        for  join in joins:
            res += join[0] + ' -> ' + join[1]
            if len(join) == 3:
                res += " " + join[2]
            res +=';\n'
        res += "}\n}"

        fname = out_file + ".dot"
        _f = open (fname, 'w')
        _f.write(res)
        _f.flush()
        _f.close()

        if dont_run_dot:
           print fname
        else:
            dot2png(fname,self.shape)
            print fname.rstrip(".dot") + ".png"

        # make and print families as json to show the packages in each family
        res = ["{"]
        idx = 0
        for f in fams:
            idx += 1
            line = "'%s': %s, " % (f, fams[f])
            line = line.replace("'", '"')
            if idx == len(fams):
                line = line[:len(line)-2]
            res.append(line + '\n')
        res.append("}")

        fname = out_file + ".FAMILIES.json"
        _f = open (fname, 'w')
        for line in res:
            _f.write(line)
        _f.flush()
        _f.close()

if __name__ == "__main__":

    parser = optparse.OptionParser()
    parser.add_option("-r", "--root",
                      type="string",
                      dest="root",
                      default="",
                      help="The root node. Pass it to prevent summarizing the"\
                        " root node into a family."
                     )
    parser.add_option("-o", "--out-file",
                      type="string",
                      dest="out_file",
                      default="out",
                      help="The generated  dot filename without the '.dot'"\
                        " extension."
                     )
    parser.add_option("-d", "--dot-file",
                      type="string",
                      dest="path",
                      default="",
                      help="The path to the dot file"
                     )
    parser.add_option("-s", "--summarize",
                      dest="summarize",
                      action="store_true",
                      default=False,
                      help="Produce dot with families summarized as single "\
                        "nodes"
                     )
    parser.add_option("-l", "--get-popular-lnodes",
                      dest="get_pop_lnodes",
                      action="store_true",
                      default=False,
                      help="Output lnodes that point to at least 'popularity'"\
                        " rnodes"
                     )
    parser.add_option("-c", "--get-popular-rnodes",
                      dest="get_pop_rnodes",
                      action="store_true",
                      default=False,
                      help="Output rnodes that point to at least 'popularity'"\
                        " lnodes"
                     )
    parser.add_option("-p", "--popularity",
                      type="string",
                      dest="popularity",
                      default="",
                      help="The minimum poplarity needed to be reported when"\
                        " -l or -r are used."
                     )
    parser.add_option("-n", "--dont-run-dot",
                      dest="dont_run_dot",
                      action="store_true",
                      default=False,
                      help="Execute dot to process the dot file to a png. "\
                        "Note very large dot files produced in rdepends mode"\
                        " for packages with many rdepends may take an "\
                        "unreasonable amount of time for dot to process."
                    )
    parser.add_option("-N", "--node-shape",
                      type="string",
                      dest="shape",
                      default="",
                      help="The default node shape in the generated png."
                    )
    parser.add_option("-v", "--verbose",
                      dest="verbose",
                      action="store_true",
                      default=False,
                      help="Verbose mode."
                    )
    (opt, args) = parser.parse_args()
    dot = Dot(opt.path, opt.root)
    if opt.summarize:
        dot.summarize(opt.out_file)
    elif opt.get_pop_lnodes:
        lpopular = dot.get_popular_lnodes(opt.popularity)
        for l in lpopular: print "%s: %d" % (l, lpopular[l])

    sys.exit()
