#!/usr/bin/python
# -*- Mode: python; indent-tabs-mode: nil; tab-width: 4 -*- ## emacs weenies

# Copyright (C) 2012 Canonical, Ltd

# Authors:
#   Michael Carifio <Michael.Carifio@canonical.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
test_dotdepends_executable -- tests dotdepends

Unittests for the main program dotdepends (/usr/bin/dotdepends when installed).

usage: python /path/to/test_dotdepends_executable.py
       nosetests -v --exe -d /path/to/test ## run all tests, including this one
"""




import sys
import os
import subprocess
import logging
import tempfile
import shutil
import unittest
logging.basicConfig(level = logging.DEBUG) ## TODO: how can the test runner pass in args?

# here: is the absolute path of where this test suite is running from.
# Find 'here' now *before* you modify sys.path. Otherwise python reports
# relative paths.
here = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
logging.debug(here)

# Point at the objects under test.
sys.path.insert(0,
    os.path.abspath(os.path.dirname(os.path.abspath(__file__))
    + '/..'))

# Point at (possible) supporting test code.
sys.path.insert( 0, os.path.dirname(__file__ ))


class DotdependsExecutableTestcase(unittest.TestCase):
    """DotdependsExecutableTestcase

    Run dotdepends on certain packages and with certain switches. This class needs more work.

    """

    # working_directory is where dotdepends makes files
    working_directory = tempfile.mkdtemp()
    good_lower_pkg = 'debconf'
    good_upper_pkg = 'gedit'
    bad_package = 'ima_baddy'
    dotdepends = here + '/../usr/bin/dotdepends'

    def dotdepends_path(self):
        return self.dotdepends


    @classmethod
    def setUpClass(cls):
        # Move to the working directory.
        os.chdir(cls.working_directory)
        logging.debug("current directory: %s" % cls.working_directory)


    @classmethod
    def tearDownClass(cls):
        # TODO: remove directory iff all tests pass. Currently removed regardless of test results.
        shutil.rmtree(cls.working_directory)



    # Name must be test_<something> so test runner can find them.
    def test_always_succeeds(self):
        """Always succeeds. Tests the test runner."""
        self.assertTrue(True, "tests the test runner")


    def test_can_find_executable(self):
        """Is dotdepends where we expect it to be?"""
        logging.debug("dotdepends: " + self.dotdepends_path())
        self.assertTrue(os.path.exists(self.dotdepends_path()))

    def test_dotdepends_installed(self):
        """dotdepends -i good_package => 1"""
        command = [self.dotdepends_path(), '-i', self.good_lower_pkg ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 1)
        self.assertTrue(stderr.startswith('Error.'))
        self.assertEqual(len(stdout), 0)


    def test_dotdepends_bad_switch(self):
        """dotdepends --invalid-switch good_pkg => 2"""
        command = [ self.dotdepends_path(), '--invalid-switch', self.good_lower_pkg ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 2)
        self.assertEqual(len(stdout), 0)

    def test_dotdepends_mode_depends_switch_i(self):
        command = [self.dotdepends_path(), '--installed', '-m', 'depends', self.good_lower_pkg ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 1)
        self.assertEqual(len(stdout), 0)

    def test_dotdepends_mode_depends(self):
        command = [self.dotdepends_path(), '-m', 'depends', self.good_lower_pkg ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 0)
        self.assertGreater(len(stdout), 0)


    def test_dotdepends_analyze(self):
        command = [ self.dotdepends_path(), self.good_lower_pkg ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 0)
        self.assertGreater(len(stdout), 0)


    def test_dotdepends_bad_package(self):
        command = [ self.dotdepends_path(), self.bad_package ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 1)
        self.assertEqual(len(stdout), 0)
        self.assertFalse(os.path.exists(self.working_directory + os.path.sep + self.bad_package ))


    def test_dotdepends_rdepends(self):
        command = [ self.dotdepends_path(), '-m', 'rdepends', self.good_upper_pkg ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 0)
        self.assertGreater(len(stdout), 0)

    def test_dotdepends_rdepends_installed_switch(self):
        command = [ self.dotdepends_path(), '-m', 'rdepends', '-i', self.good_upper_pkg ]
        p = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        stdout, stderr = p.communicate()
        self.assertEquals(p.returncode, 0)
        self.assertGreater(len(stdout), 0)



# This sequence lets you run the test suite with python directly rather
# than with a test runner indirectly.
def main():
    unittest.main()

if __name__ == '__main__':
    main()
